CREATE OR REPLACE PACKAGE BODY trainingmgr.xx_plan_app
IS

  -- Creates a new application type
  FUNCTION create_application_type (
    p_title       IN VARCHAR2
  , p_price       IN NUMBER
  , p_description IN VARCHAR2
  ) RETURN NUMBER
  IS
    
    l_application_type_id NUMBER;
    
  BEGIN
  
    SELECT xx_plan_app_types_seq.NEXTVAL
    INTO l_application_type_id 
    FROM dual;
  
    INSERT INTO xx_plan_app_types (
      id
    , start_datetime
    , end_datetime
    , title
    , price
    , description
    )
    VALUES (
      l_application_type_id
    , SYSDATE
    , NULL
    , p_title
    , p_price
    , p_description
    );
    
    RETURN l_application_type_id;
    
  END create_application_type;
  
  -- Create planning application reference from application id.
  FUNCTION create_application_reference (
    p_application_id IN NUMBER
  ) RETURN VARCHAR2
  IS
  BEGIN
    RETURN 'PLAN/' || TO_CHAR(SYSDATE, 'YYYY/MM/') || p_application_id;
  END create_application_reference;

  -- Creates a new draft application  
  FUNCTION create_new_application (
    p_application_type_id IN NUMBER
  ) RETURN NUMBER
  IS
  
    l_application_id NUMBER;
    l_application_detail_id NUMBER;
    
  BEGIN
  
    SELECT
      xx_plan_apps_seq.NEXTVAL 
    , xx_plan_app_details_seq.NEXTVAL
    INTO
      l_application_id 
    , l_application_detail_id
    FROM dual;
        
    INSERT INTO xx_plan_apps (
      id
    , type_id
    , application_reference
    )
    VALUES (
      l_application_id
    , p_application_type_id
    , create_application_reference(l_application_id)
    );
    
    INSERT INTO xx_plan_app_details (
      id
    , application_id
    , start_datetime
    , status_control
    , status
    , xml_data
    )
    VALUES (
      l_application_detail_id
    , l_application_id
    , SYSDATE
    , 'C'
    , 'DRAFT'
    , XMLELEMENT("APPLICATION_DETAIL")
    );

    RETURN l_application_id;
  
  END create_new_application;
  
  -- Updates an existing application data.
  FUNCTION update_application_data (
    p_application_id NUMBER
  , p_xml_data XMLTYPE
  ) RETURN NUMBER
  IS
  
    l_status VARCHAR2(40);
    l_application_detail_id NUMBER;
      
  BEGIN 
  
    UPDATE trainingmgr.xx_plan_app_details pad
    SET pad.status_control = NULL
    , pad.end_datetime = SYSDATE
    WHERE pad.status_control = 'C'
    AND pad.end_datetime IS NULL
    AND pad.application_id = p_application_id
    RETURNING pad.status
    INTO l_status;
    
    SELECT trainingmgr.xx_plan_app_details_seq.NEXTVAL
    INTO l_application_detail_id
    FROM dual;

    INSERT INTO trainingmgr.xx_plan_app_details (
      id
    , application_id
    , start_datetime
    , status_control
    , status
    , xml_data
    )
    VALUES (
      l_application_detail_id
    , p_application_id
    , SYSDATE
    , 'C'
    , l_status
    , p_xml_data
    );
    
    RETURN l_application_detail_id;    
    
  END update_application_data;
  
  PROCEDURE update_application_status (
    p_application_id IN NUMBER
  , p_new_status     IN VARCHAR2
  )
  IS
  
    l_lock NUMBER;
    l_sysdate DATE := SYSDATE; -- To ensure consistency throughout
    l_xml_data XMLTYPE;
    
  BEGIN
  
    /* lock application to ensure no one else can attempt update */
    SELECT id
    INTO l_lock
    FROM trainingmgr.xx_plan_apps pa
    WHERE pa.id = p_application_id
    FOR UPDATE NOWAIT;
    
    SELECT pad.xml_data
    INTO l_xml_data
    FROM trainingmgr.xx_plan_app_details pad
    WHERE pad.application_id = p_application_id
    AND pad.status_control = 'C';
    
    UPDATE trainingmgr.xx_plan_app_details pad
    SET pad.end_datetime = l_sysdate
    , pad.status_control = NULL
    WHERE pad.application_id = p_application_id
    AND pad.status_control = 'C';
    
    INSERT INTO trainingmgr.xx_plan_app_details (
      id
    , application_id
    , start_datetime
    , end_datetime
    , status_control
    , status
    , xml_data
    ) VALUES (
      trainingmgr.xx_plan_app_details_seq.NEXTVAL
    , p_application_id
    , l_sysdate
    , NULL
    , 'C'
    , p_new_status
    , l_xml_data
    );
  
  END update_application_status;
  
  FUNCTION get_unassigned_reviewer (
    p_application_uref IN VARCHAR2  
  ) RETURN NUMBER
  IS
  
    l_person_id NUMBER;
  
  BEGIN
  
    SELECT MIN(rmc.person_id)
    INTO l_person_id
    FROM decmgr.resource_members_current rmc
    WHERE rmc.res_type = 'XX_TRAINING_PLAN_APP_TEAM'
    AND rmc.role_name = 'PLAN_APP_REVIEWER'
    AND rmc.person_id || 'PER' NOT IN (
      SELECT xbc.secondary_data_uref
      FROM bpmmgr.xview_business_contexts xbc
      JOIN bpmmgr.business_routine_contexts brc ON xbc.bc_id = brc.bc_id
      JOIN bpmmgr.business_stages bs ON brc.id = bs.brc_id
      WHERE xbc.primary_data_uref = p_application_uref
      AND bs.end_datetime IS NULL
      AND bs.stage_label = 'PLNAPPRVW3'
    );
    
    RETURN l_person_id;
  
  END get_unassigned_reviewer;
  
  FUNCTION count_assigned_reviewers (
    p_application_uref IN VARCHAR2
  ) RETURN NUMBER
  IS
  
    l_count NUMBER;
    
  BEGIN
  
    SELECT COUNT(rmc.person_id)
    INTO l_count
    FROM decmgr.resource_members_current rmc
    WHERE rmc.res_type = 'XX_TRAINING_PLAN_APP_TEAM'
    AND rmc.role_name = 'PLAN_APP_REVIEWER'
    AND rmc.person_id || 'PER' IN (
      SELECT xbc.secondary_data_uref
      FROM bpmmgr.xview_business_contexts xbc
      JOIN bpmmgr.business_routine_contexts brc ON xbc.bc_id = brc.bc_id
      JOIN bpmmgr.business_stages bs ON brc.id = bs.brc_id
      WHERE xbc.primary_data_uref = p_application_uref
      AND bs.end_datetime IS NULL
      AND bs.stage_label = 'PLNAPPRVW3'
    );
    
    RETURN l_count;
    
  END count_assigned_reviewers;
  
  FUNCTION count_unassigned_reviewers (
    p_application_uref IN VARCHAR2
  ) RETURN NUMBER
  IS
  
    l_count NUMBER;
    
  BEGIN
  
    SELECT COUNT(rmc.person_id)
    INTO l_count
    FROM decmgr.resource_members_current rmc
    WHERE rmc.res_type = 'XX_TRAINING_PLAN_APP_TEAM'
    AND rmc.role_name = 'PLAN_APP_REVIEWER'
    AND rmc.person_id || 'PER' NOT IN (
      SELECT xbc.secondary_data_uref
      FROM bpmmgr.xview_business_contexts xbc
      JOIN bpmmgr.business_routine_contexts brc ON xbc.bc_id = brc.bc_id
      JOIN bpmmgr.business_stages bs ON brc.id = bs.brc_id
      WHERE xbc.primary_data_uref = p_application_uref
      AND bs.end_datetime IS NULL
      AND bs.stage_label = 'PLNAPPRVW3'
    );
    
    RETURN l_count;
    
  END count_unassigned_reviewers;
  
  FUNCTION review_manage_action_desc (
    p_application_uref IN VARCHAR2
  ) RETURN VARCHAR2
  IS
  BEGIN
  
    RETURN CHR(10) || count_assigned_reviewers(p_application_uref) || ' Reviewers Assigned' || CHR(10)
      || count_unassigned_reviewers(p_application_uref) || ' Reviewers Not Assigned';
    
  END review_manage_action_desc;
  
  FUNCTION get_application_status (
    p_application_id IN NUMBER
  ) RETURN VARCHAR2
  IS
  
    l_status VARCHAR2(50);
    
  BEGIN
  
    SELECT pad.status
    INTO l_status
    FROM trainingmgr.xx_plan_app_details pad
    WHERE pad.application_id = p_application_id
    AND pad.status_control = 'C';
    
    RETURN l_status;
    
  END get_application_status;
  
END xx_plan_app;
/
